@@ -1,7 +1,7 @@
-/* $Id: isakmp_xauth.c,v 1.14 2004/11/30 00:46:09 manubsd Exp $ */
+/* $Id: isakmp_xauth.c,v 1.17.2.5 2005/05/20 07:31:09 manubsd Exp $ */
 
 /*
- * Copyright (C) 2004 Emmanuel Dreyfus
+ * Copyright (C) 2004-2005 Emmanuel Dreyfus
  * All rights reserved.
  * 
  * Redistribution and use in source and binary forms, with or without
@@ -43,6 +43,9 @@
 #include <string.h>
 #include <errno.h>
 #include <pwd.h>
+#ifdef HAVE_SHADOW_H
+#include <shadow.h>
+#endif
 #if TIME_WITH_SYS_TIME
 # include <sys/time.h>
 # include <time.h>
@@ -70,6 +73,8 @@
 #include "crypto_openssl.h"
 #include "isakmp_var.h"
 #include "isakmp.h"
+#include "admin.h"
+#include "privsep.h"
 #include "evt.h"
 #include "handler.h"
 #include "throttle.h"
@@ -85,8 +90,22 @@
 
 #ifdef HAVE_LIBRADIUS
 #include <radlib.h>
+
+struct rad_handle *radius_auth_state = NULL;
+struct rad_handle *radius_acct_state = NULL;
 #endif
 
+#ifdef HAVE_LIBPAM
+#include <security/pam_appl.h>
+
+static char *PAM_usr = NULL;
+static char *PAM_pwd = NULL;
+static int PAM_conv(int, const struct pam_message **, 
+    struct pam_response **, void *);
+static struct pam_conv PAM_chat = { &PAM_conv, NULL };
+#endif
+
+
 void 
 xauth_sendreq(iph1)
 	struct ph1handle *iph1;
@@ -245,12 +264,18 @@ xauth_attr_reply(iph1, attr, id)
 
 		switch (isakmp_cfg_config.authsource) {
 		case ISAKMP_CFG_AUTH_SYSTEM:
-			res = xauth_login_system(iph1, usr, pwd);
+			res = privsep_xauth_login_system(usr, pwd);
 			break;
 #ifdef HAVE_LIBRADIUS
 		case ISAKMP_CFG_AUTH_RADIUS:
 			res = xauth_login_radius(iph1, usr, pwd);
 			break;
+#endif
+#ifdef HAVE_LIBPAM
+		case ISAKMP_CFG_AUTH_PAM:
+			res = privsep_xauth_login_pam(iph1->mode_cfg->port, 
+			    iph1->remote, usr, pwd);
+			break;
 #endif
 		default:
 			plog(LLV_ERROR, LOCATION, NULL, 
@@ -399,63 +424,87 @@ xauth_sendstatus(iph1, status, id)
 }
 
 #ifdef HAVE_LIBRADIUS
+int
+xauth_radius_init(void)
+{
+	/* For first time use, initialize Radius */
+	if ((isakmp_cfg_config.authsource == ISAKMP_CFG_AUTH_RADIUS) &&
+	    (radius_auth_state == NULL)) {
+		if ((radius_auth_state = rad_auth_open()) == NULL) {
+			plog(LLV_ERROR, LOCATION, NULL, 
+			    "Cannot init libradius\n");
+			return -1;
+		}
+
+		if (rad_config(radius_auth_state, NULL) != 0) {
+			plog(LLV_ERROR, LOCATION, NULL, 
+			    "Cannot open librarius config file: %s\n", 
+			    rad_strerror(radius_auth_state));
+			rad_close(radius_auth_state);
+			radius_auth_state = NULL;
+			return -1;
+		}
+	}
+
+	if ((isakmp_cfg_config.accounting == ISAKMP_CFG_ACCT_RADIUS) &&
+	    (radius_acct_state == NULL)) {
+		if ((radius_acct_state = rad_auth_open()) == NULL) {
+			plog(LLV_ERROR, LOCATION, NULL, 
+			    "Cannot init libradius\n");
+			return -1;
+		}
+
+		if (rad_config(radius_acct_state, NULL) != 0) {
+			plog(LLV_ERROR, LOCATION, NULL, 
+			    "Cannot open librarius config file: %s\n", 
+			    rad_strerror(radius_acct_state));
+			rad_close(radius_acct_state);
+			radius_acct_state = NULL;
+			return -1;
+		}
+	}
+
+	return 0;
+}
+
 int
 xauth_login_radius(iph1, usr, pwd)
 	struct ph1handle *iph1;
 	char *usr;
 	char *pwd;
 {
-	static struct rad_handle *radius_state = NULL;
 	int res;
 	const void *data;
 	size_t len;
 	int type;
 
-	/* For first time use, initialize Radius */
-	if (radius_state == NULL) {
-		if ((radius_state = rad_auth_open()) == NULL) {
-			plog(LLV_ERROR, LOCATION, NULL, 
-			    "Cannot init librradius\n");
-			return -1;
-		}
-
-		if (rad_config(radius_state, NULL) != 0) {
-			plog(LLV_ERROR, LOCATION, NULL, 
-			    "Cannot open librarius config file: %s\n", 
-			    rad_strerror(radius_state));
-			rad_close(radius_state);
-			radius_state = NULL;
-			return -1;
-		}
-	}
-
-	if (rad_create_request(radius_state, RAD_ACCESS_REQUEST) != 0) {
+	if (rad_create_request(radius_auth_state, RAD_ACCESS_REQUEST) != 0) {
 		plog(LLV_ERROR, LOCATION, NULL, 
 		    "rad_create_request failed: %s\n", 
-		    rad_strerror(radius_state));
+		    rad_strerror(radius_auth_state));
 		return -1;
 	}
 	
-	if (rad_put_string(radius_state, RAD_USER_NAME, usr) != 0) {
+	if (rad_put_string(radius_auth_state, RAD_USER_NAME, usr) != 0) {
 		plog(LLV_ERROR, LOCATION, NULL, 
 		    "rad_put_string failed: %s\n", 
-		    rad_strerror(radius_state));
+		    rad_strerror(radius_auth_state));
 		return -1;
 	}
 
-	if (rad_put_string(radius_state, RAD_USER_PASSWORD, pwd) != 0) {
+	if (rad_put_string(radius_auth_state, RAD_USER_PASSWORD, pwd) != 0) {
 		plog(LLV_ERROR, LOCATION, NULL, 
 		    "rad_put_string failed: %s\n", 
-		    rad_strerror(radius_state));
+		    rad_strerror(radius_auth_state));
 		return -1;
 	}
 
-	if (isakmp_cfg_radius_common(radius_state, iph1->mode_cfg->port) != 0)
+	if (isakmp_cfg_radius_common(radius_auth_state, iph1->mode_cfg->port) != 0)
 		return -1;
 
-	switch (res = rad_send_request(radius_state)) {
+	switch (res = rad_send_request(radius_auth_state)) {
 	case RAD_ACCESS_ACCEPT:
-		while ((type = rad_get_attr(radius_state, &data, &len)) != 0) {
+		while ((type = rad_get_attr(radius_auth_state, &data, &len)) != 0) {
 			switch (type) {
 			case RAD_FRAMED_IP_ADDRESS:
 				iph1->mode_cfg->addr4 = rad_cvt_addr(data);
@@ -486,7 +535,7 @@ xauth_login_radius(iph1, usr, pwd)
 	case -1:
 		plog(LLV_ERROR, LOCATION, NULL, 
 		    "rad_send_request failed: %s\n", 
-		    rad_strerror(radius_state));
+		    rad_strerror(radius_auth_state));
 		return -1;
 		break;
 	default:
@@ -500,26 +549,172 @@ xauth_login_radius(iph1, usr, pwd)
 }
 #endif
 
+#ifdef HAVE_LIBPAM
+static int 
+PAM_conv(msg_count, msg, rsp, dontcare)
+	int msg_count;
+	const struct pam_message **msg;
+	struct pam_response **rsp;
+	void *dontcare;
+{
+	int i;
+	int replies = 0;
+	struct pam_response *reply = NULL;
+
+	if ((reply = racoon_malloc(sizeof(*reply) * msg_count)) == NULL) 
+		return PAM_CONV_ERR;
+	bzero(reply, sizeof(*reply) * msg_count);
+
+	for (i = 0; i < msg_count; i++) {
+		switch (msg[i]->msg_style) {
+		case PAM_PROMPT_ECHO_ON:
+			/* Send the username, libpam frees resp */
+			reply[i].resp_retcode = PAM_SUCCESS;
+			reply[i].resp = strdup(PAM_usr);
+			break;
+
+		case PAM_PROMPT_ECHO_OFF:
+			/* Send the password, libpam frees resp */
+			reply[i].resp_retcode = PAM_SUCCESS;
+			reply[i].resp = strdup(PAM_pwd);
+			break;
+
+		case PAM_TEXT_INFO:
+		case PAM_ERROR_MSG:
+			reply[i].resp_retcode = PAM_SUCCESS;
+			reply[i].resp = NULL;
+			break;
+
+		default:
+			if (reply != NULL)
+				racoon_free(reply);
+			return PAM_CONV_ERR;
+			break;
+		}
+	}
+
+	if (reply != NULL)
+		*rsp = reply;
+
+	return PAM_SUCCESS;
+}
+
 int
-xauth_login_system(iph1, usr, pwd)
-	struct ph1handle *iph1;
+xauth_login_pam(port, raddr, usr, pwd)
+	int port;
+	struct sockaddr *raddr;
+	char *usr;
+	char *pwd;
+{
+	int error;
+	int res;
+	const void *data;
+	size_t len;
+	int type;
+	char *remote;
+	pam_handle_t *pam = NULL;
+
+	if (isakmp_cfg_config.port_pool == NULL) {
+		plog(LLV_ERROR, LOCATION, NULL,
+		    "isakmp_cfg_config.port_pool == NULL\n");
+		return -1;
+	}
+
+	if ((error = pam_start("racoon", usr, 
+	    &PAM_chat, &isakmp_cfg_config.port_pool[port].pam)) != 0) {
+		if (isakmp_cfg_config.port_pool[port].pam == NULL) {
+			plog(LLV_ERROR, LOCATION, NULL, "pam_start failed\n");
+			return -1;
+		} else {
+			plog(LLV_ERROR, LOCATION, NULL, 
+			    "pam_start failed: %s\n", 
+			    pam_strerror(isakmp_cfg_config.port_pool[port].pam,
+			    error));
+			goto out;
+		}
+	}
+	pam = isakmp_cfg_config.port_pool[port].pam;
+
+	if ((remote = strdup(saddrwop2str(raddr))) == NULL) {
+		plog(LLV_ERROR, LOCATION, NULL,
+		    "cannot allocate memory: %s\n", strerror(errno)); 
+		goto out;
+	}
+	
+	if ((error = pam_set_item(pam, PAM_RHOST, remote)) != 0) {
+		plog(LLV_ERROR, LOCATION, NULL, 
+		    "pam_set_item failed: %s\n", 
+		    pam_strerror(pam, error));
+		goto out;
+	}
+
+	PAM_usr = usr;
+	PAM_pwd = pwd;
+	error = pam_authenticate(pam, 0);
+	PAM_usr = NULL;
+	PAM_pwd = NULL;
+	if (error != 0) {
+		plog(LLV_ERROR, LOCATION, NULL, 
+		    "pam_authenticate failed: %s\n", 
+		    pam_strerror(pam, error));
+		goto out;
+	}
+
+	if ((error = pam_acct_mgmt(pam, 0)) != 0) {
+		plog(LLV_ERROR, LOCATION, NULL, 
+		    "pam_acct_mgmt failed: %s\n", 
+		    pam_strerror(pam, error));
+		goto out;
+	}
+
+	if ((error = pam_setcred(pam, 0)) != 0) {
+		plog(LLV_ERROR, LOCATION, NULL, 
+		    "pam_setcred failed: %s\n", 
+		    pam_strerror(pam, error));
+		goto out;
+	}
+
+	return 0;
+
+out:
+	pam_end(pam, error);
+	isakmp_cfg_config.port_pool[port].pam = NULL;
+	return -1;
+}
+#endif
+
+int
+xauth_login_system(usr, pwd)
 	char *usr;
 	char *pwd;
 {
 	struct passwd *pw;
 	char *cryptpwd;
+	char *syscryptpwd;
+#ifdef HAVE_SHADOW_H
+	struct spwd *spw;
+
+	if ((spw = getspnam(usr)) == NULL)
+		return -1;
+
+	syscryptpwd = spw->sp_pwdp;
+#endif
 
 	if ((pw = getpwnam(usr)) == NULL)
 		return -1;
 
+#ifndef HAVE_SHADOW_H
+	syscryptpwd = pw->pw_passwd;
+#endif
+
 	/* No root login. Ever. */
 	if (pw->pw_uid == 0)
 		return -1;
 
-	if ((cryptpwd = crypt(pwd, pw->pw_passwd)) == NULL)
+	if ((cryptpwd = crypt(pwd, syscryptpwd)) == NULL)
 		return -1;
 
-	if (strcmp(cryptpwd, pw->pw_passwd) == 0)
+	if (strcmp(cryptpwd, syscryptpwd) == 0)
 		return 0;
 
 	return -1;
@@ -734,7 +929,7 @@ isakmp_xauth_set(iph1, attr)
 			EVT_PUSH(iph1->local, iph1->remote, 
 			    EVTT_XAUTH_FAILED, NULL);
 
-			iph1->mode_cfg->flags &= ISAKMP_CFG_DELETE_PH1;
+			iph1->mode_cfg->flags |= ISAKMP_CFG_DELETE_PH1;
 		} else {
 			EVT_PUSH(iph1->local, 
 			    iph1->remote, EVTT_XAUTH_SUCCESS, NULL);
@@ -793,3 +988,4 @@ xauth_rmstate(xst)
 
 	return;
 }
+
