@@ -1,4 +1,4 @@
-/* $Id: pfkey.c,v 1.31.2.5 2005/05/03 20:47:24 manubsd Exp $ */
+/* $Id: pfkey.c,v 1.31.2.10 2005/10/03 14:52:19 manubsd Exp $ */
 
 /*
  * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
@@ -154,6 +154,10 @@ NULL,	/* SADB_X_SPDSETIDX */
 pk_recvspdexpire,
 NULL,	/* SADB_X_SPDDELETE2 */
 NULL,	/* SADB_X_NAT_T_NEW_MAPPING */
+NULL, /* SADB_X_MIGRATE */
+#if (SADB_MAX > 24)
+#error "SADB extra message?"
+#endif
 };
 
 static int addnewsp __P((caddr_t *));
@@ -445,6 +449,24 @@ ipsecdoi2pfkey_aalg(hashtype)
 		return SADB_AALG_MD5HMAC;
 	case IPSECDOI_ATTR_AUTH_HMAC_SHA1:
 		return SADB_AALG_SHA1HMAC;
+	case IPSECDOI_ATTR_AUTH_HMAC_SHA2_256:
+#if (defined SADB_X_AALG_SHA2_256) && !defined(SADB_X_AALG_SHA2_256HMAC)
+		return SADB_X_AALG_SHA2_256;
+#else
+		return SADB_X_AALG_SHA2_256HMAC;
+#endif
+	case IPSECDOI_ATTR_AUTH_HMAC_SHA2_384:
+#if (defined SADB_X_AALG_SHA2_384) && !defined(SADB_X_AALG_SHA2_384HMAC)
+		return SADB_X_AALG_SHA2_384;
+#else
+		return SADB_X_AALG_SHA2_384HMAC;
+#endif
+	case IPSECDOI_ATTR_AUTH_HMAC_SHA2_512:
+#if (defined SADB_X_AALG_SHA2_512) && !defined(SADB_X_AALG_SHA2_512HMAC)
+		return SADB_X_AALG_SHA2_512;
+#else
+		return SADB_X_AALG_SHA2_512HMAC;
+#endif
 	case IPSECDOI_ATTR_AUTH_KPDK:		/* need special care */
 		return SADB_AALG_NONE;
 
@@ -838,8 +860,8 @@ pk_sendgetspi(iph2)
 		/* this works around a bug in Linux kernel where it allocates 4 byte
 		   spi's for IPCOMP */
 		else if (satype == SADB_X_SATYPE_IPCOMP) {
-			minspi = ntohl (0x100);
-			maxspi = ntohl (0xffff);
+			minspi = 0x100;
+			maxspi = 0xffff;
 		}
 		else {
 			minspi = 0;
@@ -1057,9 +1079,9 @@ pk_sendupdate(iph2)
 			natt.dport = extract_port (iph2->ph1->local);
 			natt.oa = NULL;		// FIXME: Here comes OA!!!
 			natt.frag = iph2->ph1->rmconf->esp_frag;
-		}
-		else
+		} else {
 			memset (&natt, 0, sizeof (natt));
+		}
 
 		if (pfkey_send_update_nat(
 				lcconf->sock_pfkey,
@@ -1350,10 +1372,14 @@ pk_sendadd(iph2)
 			natt.dport = extract_port (iph2->ph1->remote);
 			natt.oa = NULL;		// FIXME: Here comes OA!!!
 			natt.frag = iph2->ph1->rmconf->esp_frag;
-		}
-		else
+		} else {
 			memset (&natt, 0, sizeof (natt));
 
+			/* Remove port information, that SA doesn't use it */
+			set_port(src, 0);
+			set_port(dst, 0);
+		}
+
 		if (pfkey_send_add_nat(
 				lcconf->sock_pfkey,
 				satype,
@@ -1377,6 +1403,10 @@ pk_sendadd(iph2)
 #else
 		plog(LLV_DEBUG, LOCATION, NULL, "call pfkey_send_add\n");
 
+		/* Remove port information, it is not used without NAT-T */
+		set_port(src, 0);
+		set_port(dst, 0);
+
 		if (pfkey_send_add(
 				lcconf->sock_pfkey,
 				satype,
@@ -2121,7 +2151,7 @@ pk_recvspdupdate(mhp)
 	sp = getsp(&spidx);
 	if (sp == NULL) {
 		plog(LLV_ERROR, LOCATION, NULL,
-			"such policy does not already exist: %s\n",
+			"such policy does not already exist: \"%s\"\n",
 			spidx2str(&spidx));
 	} else {
 		remsp(sp);
